#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from enum import Enum
import logging

from gcl_iam import exceptions
from gcl_iam import rules


LOG = logging.getLogger(__name__)


class OrderedEnum(Enum):

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class BasicOrderedGrant(OrderedEnum):
    def __bool__(self):
        if self.value > 0:
            return True
        return False


class Grant(BasicOrderedGrant):

    DENY = 0
    # VIEWER = 1
    # REGULAR = 10
    ALLOW = 100


class PermissionLevel(collections.defaultdict):
    def get(self, key, default=None):
        if res := super().get("*", None):
            return res
        return super().get(key, default)

    def __getitem__(self, key):
        if res := super().get("*", None):
            return res
        return super().__getitem__(key)


class BasicPermission(set):
    def get_grant_level(self, name):
        if "*" in self:
            return Grant.ALLOW
        if name in self:
            return Grant.ALLOW
        return Grant.DENY


class Enforcer(object):
    """
    A class to enforce permissions based on a list of predefined permissions.

    This class is used to check whether a given action is authorized or not.
    It allows for flexible permission definitions and supports wildcard-based
    permissions.

    Attributes:
        perms (list): A list of strings defining the permissions. Each string
                      represents a permission in the format
                      "service.resource.action".
                      Wildcards are supported, such as "*", which can be used
                      to grant all actions for a resource.

    Examples:
        # Define permissions
        perms = [
            "service.resource.action",
            "genesis_core.vm.create",
            "genesis_core.vm.*",
            "*.*.*"  # It's equal to full admin without project filters!
        ]

        # Create an Enforcer instance with the defined permissions
        enforcer = Enforcer(perms)

        # Check if a permission is granted for a specific action
        result = enforcer.enforce("genesis_core.vm.create")
        print(result)  # Output: Grant.ALLOW

        # Check if a wildcard permission applies to a resource
        result = enforcer.enforce("genesis_core.*.action")
        print(result)  # Output: Grant.ALLOW
    """

    def __init__(
        self,
        perms,
        perm_class=BasicPermission,
        level_class=PermissionLevel,
    ):
        self._perms = level_class(lambda: level_class(perm_class))
        self._load_perms(perms)

    def _load_perms(self, perms):
        for p in perms:
            service, res, perm = p.split(".", maxsplit=2)
            # Add the rule to a list of perms
            self._perms[service][res].add(perm)

    def enforce_raw(self, rule, do_raise=False, exc=None):
        rule_obj = rules.Rule.from_raw(rule)
        return self.enforce(rule_obj, do_raise, exc)

    def enforce(self, rule, do_raise=False, exc=None):
        result = Grant.DENY
        if resource := self._perms.get(rule.service):
            if permission := resource.get(rule.res):
                result = permission.get_grant_level(rule.perm)

        if do_raise and not result:
            if exc:
                raise exc(rule=(".".join((rule.service, rule.res, rule.perm))))

            raise exceptions.PolicyNotAuthorized(
                rule=(".".join((rule.service, rule.res, rule.perm)))
            )

        return result



# Copyright 2025 Genesis Corporation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import uuid


class BaseToken:

    def __init__(
        self,
        token,
        algorithm,
        audience=None,
        ignore_audience=False,
        ignore_expiration=False,
        verify=True,
    ):
        super().__init__()
        self._token = token
        self._token_info = algorithm.decode(
            token,
            audience=audience,
            ignore_audience=ignore_audience,
            ignore_expiration=ignore_expiration,
            verify=verify,
        )
        self._algorithm = algorithm

    @property
    def expiration_datetime(self):
        exp = self._token_info["exp"]
        return datetime.datetime.fromtimestamp(exp, tz=datetime.timezone.utc)

    @property
    def created_at(self):
        iat = self._token_info["iat"]
        return datetime.datetime.fromtimestamp(iat, tz=datetime.timezone.utc)

    @property
    def uuid(self):
        return uuid.UUID(self._token_info["jti"])

    @property
    def issuer_url(self):
        return self._token_info["iss"]

    @property
    def audience_name(self):
        return self._token_info["aud"]

    @property
    def user_uuid(self):
        return uuid.UUID(self._token_info["sub"])

    @property
    def token(self):
        return self._token


class AuthToken(BaseToken):

    def __init__(
        self,
        token,
        algorithm,
        audience=None,
        ignore_audience=False,
        ignore_expiration=False,
        verify=True,
    ):
        super().__init__(
            token=token,
            algorithm=algorithm,
            audience=audience,
            ignore_audience=ignore_audience,
            ignore_expiration=ignore_expiration,
            verify=verify,
        )

    @property
    def autenticated_at(self):
        auth_time = self._token_info["auth_time"]
        return datetime.datetime.fromtimestamp(
            auth_time, tz=datetime.timezone.utc
        )

    @property
    def token_type(self):
        return self._token_info["typ"]


class IdToken(BaseToken):

    def __init__(
        self,
        token,
        algorithm,
        audience=None,
        ignore_audience=False,
        ignore_expiration=False,
        verify=True,
    ):
        super().__init__(
            token=token,
            algorithm=algorithm,
            audience=audience,
            ignore_audience=ignore_audience,
            ignore_expiration=ignore_expiration,
            verify=verify,
        )

    @property
    def autenticated_at(self):
        auth_time = self._token_info["auth_time"]
        return datetime.datetime.fromtimestamp(
            auth_time, tz=datetime.timezone.utc
        )

    @property
    def user_name(self):
        return self._token_info["name"]

    @property
    def user_email(self):
        return self._token_info["email"]


class RefreshToken(BaseToken):

    def __init__(
        self,
        token,
        algorithm,
        audience=None,
        ignore_audience=False,
        ignore_expiration=False,
        verify=True,
    ):
        super().__init__(
            token=token,
            algorithm=algorithm,
            audience=audience,
            ignore_audience=ignore_audience,
            ignore_expiration=ignore_expiration,
            verify=verify,
        )



# Copyright 2025 Genesis Corporation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import jwt

from gcl_iam import constants as c


class AbstractAlgorithm(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def decode(self, data):
        raise NotImplementedError("Not implemented")

    @abc.abstractmethod
    def encode(self, data):
        raise NotImplementedError("Not implemented")


class HS256(AbstractAlgorithm):

    def __init__(self, key):
        super().__init__()
        self._key = key

    def decode(
        self,
        data,
        audience=None,
        ignore_audience=False,
        ignore_expiration=False,
        verify=True,
    ):
        options = {
            "verify_exp": not ignore_expiration,
            "verify_aud": not ignore_audience,
        }
        return jwt.decode(
            data,
            key=self._key,
            algorithms=c.ALGORITHM_HS256,
            options=options,
            audience=audience,
            verify=verify,
        )

    def encode(self, data):
        return jwt.encode(data, key=self._key, algorithm=c.ALGORITHM_HS256)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pytest
import threading
import uuid

from unittest import mock

from restalchemy.common import contexts

from gcl_iam import controllers
from gcl_iam import engines
from gcl_iam import exceptions


FAKE_PROJECT_ID = uuid.UUID("fbe1fc09-e4cc-4cd2-a51d-c823b40155b2")
FAKE_PROJECT_ID_2 = uuid.UUID("29885802-c8b9-42d5-806f-6ecc1c943bbb")
FAKE_METHOD = "fake_method"


class FakeController(controllers.PolicyBasedControllerMixin):
    __policy_service_name__ = "service"
    __policy_name__ = "vm"

    def __init__(self, *args, **kwargs):
        self._req = mock.Mock()
        super().__init__(*args, **kwargs)


@pytest.fixture
def user_context():
    ctx_storage = mock.Mock()
    contexts.ContextWithStorage._store_context_session(ctx_storage)
    ctx_storage.iam_context.introspection_info.return_value = {
        "user_info": {},
        "project_id": FAKE_PROJECT_ID,
        "otp_verified": True,
        "permission_hash": "xxxx",
        "permissions": [
            "service.resource.action",
            "genesis_core.vm.create",
            "genesis_core.vm.admin",
        ],
    }

    yield ctx_storage.iam_context.introspection_info

    contexts.ContextWithStorage._clear_context()


@pytest.fixture
def unscoped_context():
    ctx_storage = mock.Mock()
    contexts.ContextWithStorage._store_context_session(ctx_storage)
    ctx_storage.iam_context.introspection_info.return_value = {
        "user_info": {},
        "project_id": None,
        "otp_verified": True,
        "permission_hash": "xxxx",
        "permissions": [
            "service.resource.action",
            "genesis_core.vm.create",
            "genesis_core.vm.admin",
            "*.*.*",
        ],
    }

    yield ctx_storage.iam_context.introspection_info

    contexts.ContextWithStorage._clear_context()


class TestPolicyBasedControllerMixin:
    def test_default_init_with_ctx_and_project(self, user_context):
        pc = FakeController()

        assert pc._introspection is not None
        assert pc._ctx_project_id == FAKE_PROJECT_ID

    def test_auth_unscoped(self, unscoped_context):
        pc = FakeController()

        assert (
            pc._enforce_and_authorize_project_id("create", FAKE_PROJECT_ID)
            is None
        )
        assert (
            pc._enforce_and_authorize_project_id("create", FAKE_PROJECT_ID_2)
            is None
        )

    def test_auth_project_scoped_any_permission(self, unscoped_context):
        unscoped_context.return_value = {
            "user_info": {},
            "project_id": FAKE_PROJECT_ID,
            "otp_verified": True,
            "permission_hash": "xxxx",
            "permissions": [
                "service.resource.action",
                "genesis_core.vm.create",
                "genesis_core.vm.admin",
                "*.*.*",
            ],
        }

        pc = FakeController()

        assert (
            pc._enforce_and_authorize_project_id("create", FAKE_PROJECT_ID)
            is None
        )

        with pytest.raises(exceptions.Forbidden):
            pc._enforce_and_authorize_project_id("create", FAKE_PROJECT_ID_2)

    def test_auth_project_id_allowed(self, user_context):
        pc = FakeController()

        pc._enforce_and_authorize_project_id("create", FAKE_PROJECT_ID)

    def test_auth_project_id_method_forbidden(self, user_context):
        with pytest.raises(exceptions.Forbidden):
            pc = FakeController()
            pc._enforce_and_authorize_project_id(
                "strange_create", FAKE_PROJECT_ID_2
            )

    def test_auth_non_uuid_project_id_allowed(self, user_context):
        pc = FakeController()

        pc._enforce_and_authorize_project_id("create", str(FAKE_PROJECT_ID))

    def test_auth_project_id_not_allowed(self, user_context):
        with pytest.raises(exceptions.Forbidden):
            pc = FakeController()
            pc._enforce_and_authorize_project_id("create", FAKE_PROJECT_ID_2)

    def test_override_project_id_from_filters(self, user_context):
        kwargs = {"project_id": FAKE_PROJECT_ID}
        pc = FakeController()

        pc._enforce_and_override_project_id_in_kwargs("create", kwargs)

        assert kwargs == {"project_id": FAKE_PROJECT_ID}

    def test_override_project_id_from_ctx(self, user_context):
        kwargs = {}
        pc = FakeController()

        pc._enforce_and_override_project_id_in_kwargs("create", kwargs)

        assert kwargs == {"project_id": FAKE_PROJECT_ID}



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gcl_iam import enforcers
from gcl_iam import exceptions
from gcl_iam import tokens


class IamEngine:

    def __init__(self, auth_token, algorithm, driver, enforcer=None):
        super().__init__()
        self._token_info = tokens.AuthToken(
            auth_token,
            algorithm,
            ignore_audience=True,
            ignore_expiration=False,
            verify=True,
        )
        self._driver = driver
        self._introspection_info = self._driver.get_introspection_info(
            token_info=self._token_info
        )

        # Forbid requests without auth or without project scope
        if not self._introspection_info:
            raise exceptions.Unauthorized()

        self._enforcer = enforcer or enforcers.Enforcer(
            self._introspection_info["permissions"]
        )

    @property
    def token_info(self):
        return self._token_info

    def introspection_info(self):
        return self._introspection_info

    @property
    def enforcer(self):
        return self._enforcer



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Rule:
    def __init__(self, service, res, perm):
        self.service = service
        self.res = res
        self.perm = perm

    def __str__(self):
        return f"{self.service}.{self.res}.{self.perm}"

    @classmethod
    def from_raw(cls, rule):
        service, res, perm = rule.split(".", maxsplit=2)

        return cls(service, res, perm)



#    Copyright 2011 OpenStack Foundation.
#    Copyright 2020 Eugene Frolov
#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from http import client as http_client
import logging
import re

from restalchemy.api.middlewares import contexts as contexts_mw
from restalchemy.api.middlewares import errors as errors_mw

from gcl_iam import contexts
from gcl_iam import drivers
from gcl_iam import engines
from gcl_iam import exceptions as exc


LOG = logging.getLogger(__name__)


class AbstactEndpointComparator(metaclass=abc.ABCMeta):

    def _build_full_path(self, path):
        return re.compile(path)

    @abc.abstractmethod
    def compare(self, req):
        raise NotImplementedError("Not implemented")


class EndpointComparator(AbstactEndpointComparator):
    def __init__(self, path, methods=None):
        self._path = path
        self._methods = methods or ["GET"]

    def compare(self, req):
        full_path = self._build_full_path(self._path)
        return full_path.fullmatch(req.path) and req.method in self._methods


class GenesisCoreAuthMiddleware(contexts_mw.ContextMiddleware):

    def __init__(
        self,
        application,
        token_algorithm,
        context_class=contexts.GenesisCoreAuthContext,
        context_kwargs=None,
        iam_engine_driver=None,
        skip_auth_endpoints: list = None,
    ):
        super().__init__(
            application=application,
            context_class=context_class,
            context_kwargs=context_kwargs,
        )
        self._token_algorithm = token_algorithm
        self._iam_engine_driver = iam_engine_driver or drivers.HttpDriver()
        self._skip_auth_endpoints = skip_auth_endpoints or []

    def _should_skip_auth(self, req):
        for endpoint in self._skip_auth_endpoints:
            if endpoint.compare(req):
                return True
        return False

    def _get_auth_token(self, req):
        header_value = req.headers.get("Authorization", "")
        if header_value.lower().startswith("bearer "):
            return header_value.split(" ")[1]
        raise exc.InvalidAuthTokenError()

    def _get_response(self, ctx, req):
        with ctx.context_manager():
            if self._should_skip_auth(req):
                LOG.info("Skip auth for %s", req.path)
                return super()._get_response(ctx, req)
            else:
                try:
                    iam_context = engines.IamEngine(
                        auth_token=self._get_auth_token(req),
                        algorithm=self._token_algorithm,
                        driver=self._iam_engine_driver,
                    )
                except Exception:
                    LOG.exception("Invalid auth token by reason:")
                    raise exc.InvalidAuthTokenError()

                with ctx.iam_session(iam_context):
                    req.iam_engine = iam_context
                    return super()._get_response(ctx, req)


class ErrorsHandlerMiddleware(errors_mw.ErrorsHandlerMiddleware):

    forbidden_exc = (exc.CommonForbiddenError,)

    def _construct_error_response(self, req, e):
        if isinstance(e, exc.ClientAuthenticationError):
            # RFC 6749
            return req.ResponseClass(
                status=http_client.UNAUTHORIZED,
                json={
                    "error": "invalid_client",
                    "error_description": str(e),
                },
            )
        elif isinstance(e, exc.InvalidAuthTokenError):
            return req.ResponseClass(
                status=http_client.UNAUTHORIZED,
                json={
                    "error": "invalid_token",
                    "error_description": str(e),
                },
            )
        elif isinstance(e, exc.CredentialsAreInvalidError):
            # RFC 6749
            return req.ResponseClass(
                status=http_client.BAD_REQUEST,
                json={
                    "error": "invalid_grant",
                    "error_description": str(e),
                },
                headers={"WWW-Authenticate": 'Bearer error="invalid_token"'},
            )
        elif isinstance(e, self.forbidden_exc):
            return req.ResponseClass(
                status=http_client.FORBIDDEN,
                json=errors_mw.exception2dict(e),
            )
        else:
            return super()._construct_error_response(req, e)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import pytest
from gcl_iam.enforcers import Enforcer, Grant
from gcl_iam import exceptions
from gcl_iam import rules

# Sample permissions data
perms = [
    "service.resource.action",
    "genesis_core.vm.create",
    "genesis_core.vm.*",
]

admin_perm = ["service.resource.action", "*.*.*"]


def test_enforcer_init():
    # Test default initialization
    enforcer = Enforcer(perms)

    assert isinstance(enforcer._perms, collections.defaultdict)

    # Test with empty perms list
    enforcer_empty = Enforcer([])

    assert len(enforcer_empty._perms) == 0


def test_load_perms():
    enforcer = Enforcer(perms)

    expected_permissions = {
        "service": {"resource": set(["action"])},
        "genesis_core": {"vm": set(["create", "*"])},
    }

    assert enforcer._perms == expected_permissions


def test_enforce_raw():
    enforcer = Enforcer(perms)

    result = enforcer.enforce_raw("service.resource.action")

    assert result == Grant.ALLOW


def test_enforce_allow():
    enforcer = Enforcer(perms)

    result = enforcer.enforce(rules.Rule("service", "resource", "action"))

    assert result == Grant.ALLOW


def test_enforce_admin():
    enforcer = Enforcer(admin_perm)

    result = enforcer.enforce(rules.Rule("genesis_core", "resource", "mega"))

    assert result == Grant.ALLOW


def test_enforce_deny():
    enforcer = Enforcer(perms)

    result = enforcer.enforce(rules.Rule("genesis_core", "resource", "other"))

    assert result == Grant.DENY


def test_enforce_comparable_permission():
    enforcer = Enforcer(perms)

    result = enforcer.enforce_raw("genesis_core.vm.create")

    assert result >= Grant.ALLOW


def test_error_raising_on_denied_rule():
    enforcer = Enforcer(perms)

    with pytest.raises(exceptions.PolicyNotAuthorized) as excinfo:
        enforcer.enforce(
            rules.Rule("genesis_core", "resource", "other"), do_raise=True
        )

    assert "genesis_core.resource.other" in str(excinfo.value)


def test_error_raising_on_deny_rule_without_exception():
    enforcer = Enforcer(perms)

    result = enforcer.enforce(
        rules.Rule("genesis_core", "resource", "other"),
        do_raise=False,
        exc=exceptions.PolicyNotAuthorized,
    )

    assert result == Grant.DENY


def test_enforce_multiple_permissions():
    enforcer = Enforcer(perms)

    result = enforcer.enforce_raw("genesis_core.vm.*")

    assert result == Grant.ALLOW
